import regex as re


class KrautException(Exception):
    pass


class AmbiguousChord(Exception):
    pass


def note_to_number(note, shift, german=False):
    if note is None:
        return None
    if german:
        if note == "H":
            note = "B"
        elif note == "B":
            if shift == "b":
                raise AmbiguousChord
            shift = "b"
    number = ["C", 0, "D", 0, "E", "F", 0, "G", 0, "A", 0, "B"].index(note.upper())
    if shift:
        number += "b #".index(shift) - 1
    return number % 12


def normalize_modifiers(modifiers):
    return [{"+": "aug", "°": "dim"}.get(mod, mod) for mod in modifiers]


CHORD = re.compile(
    r"""^
    (?P<optional>\()?                       # opening paren, to indicate optionality
    (?P<main>[ABCDEFGHabcdefgh])            # bass chord. can already indicate minor!
    (?P<shift>[#b])?                        # shift semitone up or down
    (?P<minor>[m-])?                        # indicator for minor chord
    (?:
      \(?
      (?P<modifiers>
        sus[24]?                            # suspended chord
      | (?:dim|°)                           # diminished chord
      | (?:aug|\+)                          # augmented chord
      | (?:\#|b|no|-|[mM]aj|M|add)?(?:\d+)  # regular/b'd added tone
      | M                                   # major 7, I guess...
      )
      \)?
    )*
    (?:/
    (?P<bass>[ABCDEFGHabcdefgh])            # bass tone
    (?P<bassshift>[#b])?                    # bass shift
    )?
    \)?
    $""",
    re.VERBOSE,
)


class Chord:
    exheritables = ["transpose", "simplify"]

    @classmethod
    def is_chord(cls, token):
        return CHORD.match(token) or token == "N.C."

    def __init__(self, tone, minor=False, bass=None, modifiers=None, optional=False):
        self.tone = tone
        self.minor = minor
        self.bass = bass
        self.modifiers = modifiers or []
        self.optional = optional

    @classmethod
    def from_string(cls, string, german=False):
        if string == "N.C.":
            return cls(tone=None)
        data = CHORD.match(string).capturesdict()
        main = data["main"][0]
        shift = data["shift"][0] if data["shift"] else None
        bass = data["bass"][0] if data["bass"] else None
        bassshift = data["bassshift"][0] if data["bassshift"] else None
        if "H" in (main, bass) and not german:
            raise KrautException

        tone = note_to_number(main, shift, german=german)
        bass = note_to_number(bass, bassshift, german=german)
        minor = bool(data["minor"]) or main.islower()
        modifiers = normalize_modifiers(data["modifiers"])
        optional = bool(data["optional"])
        return cls(
            tone=tone, minor=minor, bass=bass, modifiers=modifiers, optional=optional
        )

    @staticmethod
    def note(number, flags=""):
        if "b" in flags:
            the_note = [
                "C",
                "Db",
                "D",
                "Eb",
                "E",
                "F",
                "Gb",
                "G",
                "Ab",
                "A",
                "Bb",
                "B",
            ][number]
        else:
            the_note = [
                "C",
                "C#",
                "D",
                "D#",
                "E",
                "F",
                "F#",
                "G",
                "G#",
                "A",
                "A#",
                "B",
            ][number]
        if "f" in flags:
            the_note = the_note.replace("#", "♯").replace("b", "♭")
        return the_note

    def __format__(self, code):
        return self.to_monospace(flags=code)

    def to_monospace(self, flags=""):
        if self.tone is None:
            return "N.C."
        return "".join(
            [
                self.note(self.tone, flags=flags),
                "m" if self.minor else "",
                "".join(self.modifiers),
                f"/{self.note(self.bass, flags=flags)}"
                if self.bass is not None
                else "",
            ]
        )

    def __repr__(self):
        return self.to_monospace()

    def __len__(self):
        # approximate: flags can change length
        return len(self.to_monospace())

    def length(self, flags=""):
        return len(self.to_monospace(flags=flags))

    def transpose(self, semitones):
        if self.tone is None:
            return self
        data = dict(self.__dict__)
        data["tone"] = (data["tone"] + semitones) % 12
        if data["bass"] is not None:
            data["bass"] = (data["bass"] + semitones) % 12
        return Chord(**data)

    def simplify(self):
        # TODO: several levels of simplification possible. We'll start with the
        # most extreme:
        if self.tone is None:
            return self
        return Chord(self.tone, minor=self.minor)

    def to_html(self, flags=""):
        if self.tone is None:
            return "N.C."
        if "F" not in flags:
            flags += "f"
        parts = [
            self.note(self.tone, flags=flags).replace(
                "♭", "<span class='bhack'>♭</span>"
            )
        ]
        if self.minor:
            parts.append("m")
        if self.modifiers:
            parts.append("<sup>")
            parts.append("".join(self.modifiers))
            parts.append("</sup>")
        if self.bass is not None:
            parts.append(f"/{self.note(self.bass, flags=flags)}")
        return "".join(parts)



import click

from chordy.song import Song


@click.command()
@click.argument("file", type=click.File("r"))
@click.option(
    "--type",
    "-T",
    "type_",
    type=click.Choice(["tex", "txt", "html"]),
    help="What file type chordy should produce.",
    default="txt",
)
@click.option(
    "--transpose", "-t", default=0, help="Amount of semitones to transpose up/down",
)
@click.option(
    "--simplify", "-s", is_flag=True, help="Whether to simplify the chords",
)
@click.option(
    "--flags", "-f", type=str, help="Sequence of 1-letter display codes", default=""
)
@click.option(
    "--prefer",
    "-p",
    help="Whether to prefer b/♭ or #/♯",
    type=click.Choice(["b", "#", "♯", "♭", "flat", "sharp"]),
)
def cli(file, type_, transpose, simplify, flags, prefer):
    """
    Convert a UG-style chord file into PDF/normalized chords.
    """
    song = Song.from_file(file)
    if prefer:
        flags += {"♯": "#", "♭": "b", "flat": "b", "sharp": "#"}.get(prefer, prefer)
    if transpose:
        song = song.transpose(transpose)
    if simplify:
        song = song.simplify()
    if type_ == "tex":
        print(song.to_tex(flags=flags))
    elif type_ == "html":
        print(song.to_html(flags=flags))
    else:
        print(song.to_monospace(flags=flags))


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



from io import StringIO
from textwrap import dedent

from .segments import get_segments, Newline, Span, Section
from .chord import Chord, KrautException


class Song:
    def __init__(self, title, segments):
        self.title = title
        self.segments = segments

    @classmethod
    def from_file(cls, file, name=""):
        lines = list(file)
        try:
            return cls(name, list(get_segments(lines)))
        except KrautException:
            return cls(name, list(get_segments(lines, german=True)))

    def to_monospace(self, flags=""):
        io = StringIO()

        def flush_lines(cache):
            last_chord_full_width = False
            for span in cache:
                if last_chord_full_width:
                    io.write(" ")
                if span.chord:
                    the_str = ("{:%s}" % flags).format(span.chord)
                else:
                    the_str = ""
                io.write(the_str.ljust(len(span)))
                last_chord_full_width = span.chord and len(span) == len(span.chord)

            io.write("\n")
            for span in cache:
                io.write((span.text or "").ljust(len(span)))
            io.write("\n")
            cache.clear()

        line_cache = []
        for segment in self.segments:
            if isinstance(segment, Span):
                line_cache.append(segment)
                continue
            if line_cache:
                flush_lines(line_cache)
            if not isinstance(segment, Newline):
                io.write(repr(segment))
        if line_cache:
            flush_lines(line_cache)
        return io.getvalue()

    def to_tex(self, flags=""):
        io = StringIO()
        io.write(dedent(
            r"""
            \documentclass{article}
            \usepackage{leadsheets}
            \begin{document}
            \begin{song}{title={%s}}
            """[1:] % self.title
        ))
        section = None
        for segment in self.segments:
            if isinstance(segment, Section):
                if section:
                    io.write("\\end{%s}\n" % section.envname)
                section = segment
            io.write(segment.to_tex(flags=flags))
        if section:
            io.write("\\end{%s}\n" % section.envname)
        io.write(dedent(
            r"""
            \end{song}
            \end{document}
            """.lstrip()
        ))
        return io.getvalue()

    def to_html(self, flags=""):
        return "\n".join([
            "<html><head><style>",
            """
            .span {
                line-height: 2em;
                display: inline-flex;
                flex-direction: column;
                text-align: left;
            }

            .chord {
                top: -1.8em;
            }

            .span.empty {
                margin-right: 0.5em;
            }
            .bhack {
                position: relative;
                left: -4px;
                letter-spacing: -7px;
            }
            """,
            "</style></head><body>",
            "".join(segment.to_html(flags=flags) for segment in self.segments),
            "</body></html>",
        ])

    def __getattr__(self, attr):
        if attr in Chord.exheritables:

            def fn(*args):
                data = dict(self.__dict__)
                data["segments"] = [
                    getattr(segment, attr)(*args)
                    if isinstance(segment, Span)
                    else segment
                    for segment in data["segments"]
                ]
                return Song(**data)

            return fn
        raise AttributeError(
            f"'{self.__class__.__name__} object has no attribute '{attr}'"
        )



from collections import Counter
from itertools import groupby

import regex as re

from .chord import Chord


class Segment:
    pass


class Span(Segment):
    def __init__(self, text, chord):
        self.text = text
        self.chord = chord

    def __format__(self, code):
        return ("<{:%s}: {}>" % code).format(self.chord, self.text)

    def __len__(self):
        return max(
            len(self.text) if self.text is not None else 0,
            len(self.chord) if self.chord is not None else 0,
        )

    def __getattr__(self, attr):
        if attr in Chord.exheritables:

            def fn(*args):
                if self.chord is None:
                    return self
                data = dict(self.__dict__)
                data["chord"] = getattr(self.chord, attr)(*args)
                return Span(**data)

            return fn
        raise AttributeError(
            f"'{self.__class__.__name__} object has no attribute '{attr}'"
        )

    def to_tex(self, flags=""):
        if self.chord:
            if self.text:
                return r"^{%s}%s" % (self.chord.to_monospace(flags=flags), self.text)
            return r"^{%s}\empty " % self.chord.to_monospace(flags=flags)
        return self.text

    def to_html(self, flags=""):
        return "".join(
            [
                "<span class='span{}'><span class='chord'>".format(
                    " empty" if not self.text else ""
                ),
                self.chord.to_html(flags=flags) if self.chord else "&nbsp;",
                "</span>",
                (self.text or "").replace(" ", "&nbsp;"),
                "</span>",
            ]
        )


class Newline(Segment):
    def __repr__(self):
        return "\n"

    def to_tex(self, **kwargs):
        return "\\\\\n"

    def to_html(self, **kwargs):
        return "<br>"


class Section(Segment):
    def __init__(self, title):
        self.title = title

    @property
    def envname(self):
        title = self.title.rstrip(" 1234567890:").lower()
        if title in ("verse", "chorus", "intro", "interlude", "bridge"):
            return title
        return "verse*"

    def __repr__(self):
        return f"[{self.title}]\n"
        # return "SECTION\n"

    def to_tex(self, **kwargs):
        return "\\subsection*{%s}\n" % self.envname

    def to_html(self, **kwargs):
        return f"<h3>{self.title}</h3>"


def line_replacements(line):
    tokens = line.split(" ")
    out = []
    for token in tokens:
        if out and (match := re.match(r"\(?(?:x(\d+)|(\d+)x)\)?", token)):
            number = int(next(filter(bool, match.groups())))
            out.extend([out[-1]] * (number - 1))
            continue
        out.append(token)
    return " ".join(out)


def token_type(token):
    if token.startswith("["):
        return "meta"
    if token in "Aa":
        return "ambiguous"
    if Chord.is_chord(token):
        return "chord"
    if token.isupper() and token.endswith(":"):
        return "meta"
    if token.lower().strip(":") in ["bridge", "verse", "chorus", "chords"]:
        return "meta"
    if re.match(r"^\(x\d\)$", token):
        return "instruct"
    return "word"


def line_type(line):
    tokens = [token for token in line.split(" ") if token]
    if not tokens:
        return "empty"
    types = Counter(token_type(token) for token in tokens)
    if len(types) == 2 and set(types) & {"ambiguous"}:
        types.pop("ambiguous")
    if len(types) == 2 and set(types) == {"meta", "word"}:
        types.pop("word")
    if len(types) == 1:
        type_, _ = types.popitem()
        if type_ == "ambiguous":
            type_ = "chord"
        return type_
    elif "word" in types:
        return "word"
    return "/".join(types)


def spans_from_single(line, ltype, german=False):
    if ltype == "word":
        # FIXME keep previous chord?
        # but not fixing this probably doesn't hurt
        yield Span(text=line, chord=None)
    elif ltype == "chord":
        for chord in line.split(" "):
            if not chord:
                continue
            yield Span(text=None, chord=Chord.from_string(chord, german=german))
    elif ltype == "empty":
        pass
    elif ltype == "meta":
        yield Section(line.strip().strip("[]"))
    else:
        raise NotImplementedError(line, ltype)


def spans_from_pair(cline, tline, german=False):
    spans = []
    cur_chord = None
    text = []
    for ws, group in groupby(cline, lambda x: x == " "):
        cstr = "".join(group)
        clen = len(cstr)
        if ws:
            text.extend(tline[:clen])
            tline = tline[clen:]
            spans.append(Span(text="".join(text), chord=cur_chord))
            cur_chord = None
            text = []
            continue
        cur_chord = Chord.from_string(cstr, german=german)
        text.extend(tline[:clen])
        tline = tline[clen:]
    if text or cur_chord:
        spans.append(Span(text="".join(text), chord=cur_chord))
    if tline:
        spans[-1].text += tline
    yield from spans


def get_segments(lines, german=False):
    prev = None
    for line in lines:
        line = line.rstrip("\n") + " "
        line = line_replacements(line)
        ltype = line_type(line)
        if not prev:
            prev = line, ltype
            continue
        pline, ptype = prev
        if ptype == "chord" and ltype == "word":
            yield from spans_from_pair(pline, line, german=german)
            yield Newline()
            prev = None
            continue
        yield from spans_from_single(pline, ptype, german=german)
        yield Newline()
        prev = line, ltype
    if prev is not None:
        yield from spans_from_single(*prev, german=german)
        yield Newline()
